import type { Chain, Network } from "@wormhole-foundation/sdk-base";
import type { ChainContext, Signer, TokenId } from "@wormhole-foundation/sdk-definitions";
import type { Wormhole } from "../wormhole.js";
import type { RouteTransferRequest } from "./request.js";
import type { Options, Quote, QuoteResult, Receipt, TransferParams, ValidatedTransferParams, ValidationResult } from "./types.js";
export declare abstract class Route<N extends Network, OP extends Options = Options, VP extends ValidatedTransferParams<OP> = ValidatedTransferParams<OP>, R extends Receipt = Receipt> {
    wh: Wormhole<N>;
    request: RouteTransferRequest<N>;
    abstract readonly NATIVE_GAS_DROPOFF_SUPPORTED: boolean;
    abstract readonly IS_AUTOMATIC: boolean;
    constructor(wh: Wormhole<N>, request: RouteTransferRequest<N>);
    abstract validate(params: TransferParams<OP>): Promise<ValidationResult<OP>>;
    abstract quote(params: ValidatedTransferParams<OP>): Promise<QuoteResult<OP, VP>>;
    abstract initiate(sender: Signer, quote: Quote<OP, VP>): Promise<R>;
    abstract track(receipt: R, timeout?: number): AsyncGenerator<R>;
    transferUrl(txid: string): string;
    abstract getDefaultOptions(): OP;
}
export interface RouteMeta {
    name: string;
    logo?: string;
    support?: string;
    source?: string;
}
export interface RouteConstructor {
    new <N extends Network>(wh: Wormhole<N>, request: RouteTransferRequest<N>): Route<N>;
    /**  Details about the route provided by the implementation */
    readonly meta: RouteMeta;
    /** get the list of networks this route supports */
    supportedNetworks(): Network[];
    /** get the list of chains this route supports */
    supportedChains(network: Network): Chain[];
    /** check that the underlying protocols are supported */
    isProtocolSupported<N extends Network>(chain: ChainContext<N>): boolean;
    /** get the list of source tokens that are possible to send */
    supportedSourceTokens(fromChain: ChainContext<Network>): Promise<TokenId[]>;
    /** get the list of destination tokens that may be recieved on the destination chain */
    supportedDestinationTokens<N extends Network>(token: TokenId, fromChain: ChainContext<N>, toChain: ChainContext<N>): Promise<TokenId[]>;
}
export type StaticRouteMethods<I extends RouteConstructor> = InstanceType<I>;
/**
 * AutomaticRoute is used whenever a relayer is delivering the
 * Attestation to the destination chain
 */
export declare abstract class AutomaticRoute<N extends Network, OP extends Options = Options, VP extends ValidatedTransferParams<OP> = ValidatedTransferParams<OP>, R extends Receipt = Receipt> extends Route<N, OP, VP, R> {
    IS_AUTOMATIC: boolean;
    abstract isAvailable(): Promise<boolean>;
}
export declare function isAutomatic<N extends Network>(route: Route<N>): route is AutomaticRoute<N>;
/**
 * Manual route is used whenever a manual delivery of the Attestation
 * is necessary
 */
export declare abstract class ManualRoute<N extends Network, OP extends Options = Options, VP extends ValidatedTransferParams<OP> = ValidatedTransferParams<OP>, R extends Receipt = Receipt> extends Route<N, OP, VP, R> {
    NATIVE_GAS_DROPOFF_SUPPORTED: boolean;
    IS_AUTOMATIC: boolean;
    abstract complete(sender: Signer, receipt: R): Promise<R>;
}
export declare function isManual<N extends Network>(route: Route<N>): route is ManualRoute<N>;
/**
 * FinalizableRoute is used whenever the route has a step after
 * completion that needs to be done
 */
export declare abstract class FinalizableRoute<N extends Network, OP extends Options = Options, VP extends ValidatedTransferParams<OP> = ValidatedTransferParams<OP>, R extends Receipt = Receipt> extends Route<N, OP, VP, R> {
    abstract finalize(sender: Signer, receipt: R): Promise<R>;
}
export declare function isFinalizable<N extends Network>(route: Route<N>): route is FinalizableRoute<N>;
//# sourceMappingURL=route.d.ts.map